package log

import (
	"context"
	"fmt"
	"net/http"
	"runtime/debug"
	"strconv"
	"strings"

	http2 "Crazy_Push/models/utils/http"

	"Crazy_Push/models/utils/json"

	"github.com/asim/go-micro/v3/metadata"
	"go.uber.org/zap"
)

var logger *Logger

const (
	SystemLog = "Systemlog"
)

type Logger struct {
	Url        string
	Service    string
	httpClient *http.Client
	debug      bool // 本地显示log开关
	tag        string
	level      AtomicLevel
}

const (
	GameLogModule   = "game"
	SystemLogModule = "system"
	microModule     = "micro"
)

// Log 定义了日志格式
//
// Tag 标签 对应logstash的tag和Elasticsearch和Kibana中的Index
//
// TracedId 追踪Id,追踪整个微服务链路调用的Id
//
// LogLevel 日志级别分为：Info，Warn,Error
//
// Service 服务名称
//
// Ctx 某个请求的上下文,一般里面设置了请求和响应内容。
//
// ExtField 额外字段，用于将某些常用的字段摘取出来，便于在ELK日志系统上进行系统日志分析。
type Log struct {
	Uid       int64
	LogModule string
	TracedId  interface{}
	LogLevel  string
	Service   string
	Content   string
	Code      int64
	Msg       string
	ExtField  map[string]interface{}
	Tag       string
	Cost      int64 //ms
}

func (l *Log) Level() Level {
	var lv Level
	lv.Parse(l.LogLevel)
	return lv
}

const UidKeyInMetedata = "Uid"

// SetUidToContext 往ctx设置uid
func SetUidToContext(ctx context.Context, uid int64) context.Context {
	return metadata.Set(ctx, UidKeyInMetedata, fmt.Sprintf("%d", uid))
}

func GetUidFromCtx(ctx context.Context) int64 {
	if ctx == nil {
		return 0
	}
	uidStr, ok := metadata.Get(ctx, UidKeyInMetedata)
	if !ok {
		return 0
	}
	uid, _ := strconv.ParseInt(uidStr, 10, 64)
	return uid
}

func TKLog(log *Log) {
	if logger.debug {
		ConsoleLogger.Info("TkLogInfo:", zap.Any("detail", log))
	}
	if log == nil || logger == nil {
		return
	}
	if !logger.level.Open(log.Level()) {
		return
	}
	log.Tag = logger.tag
	logStr := json.MarshalToString(log)

	_, err := logger.httpClient.Post(logger.Url, "application/json", strings.NewReader(logStr))
	if err != nil {
		ConsoleLogger.Error("TKLog failed", zap.Error(err), zap.String("Error", logStr))
	}
}

func InitLogger(url, tag, service, lv string, local bool) {
	logger = new(Logger)
	logger.Url = url
	logger.tag = tag
	logger.Service = service
	logger.httpClient = http2.CreateHTTPClient()
	logger.debug = local

	logger.level = NewAtomicLevel()
	if lv != "" {
		logger.level.SetLevelWithStr(lv)
	}

	//mLog.DefaultLogger = NewLogger()
}

func GetHttpClient() *http.Client {
	if logger == nil {
		return nil
	}
	return logger.httpClient
}

func ReloadLogger(url string) {
	logger.Url = url
}

// TkLogInfo 打印Info日志 LogModule为game
func TkLogInfo(ctx context.Context, uid int64, content string, extra ...map[string]interface{}) {
	var ex map[string]interface{}
	var traceId string
	traceId = GetTraceId(ctx)
	if uid == 0 {
		uid = GetUidFromCtx(ctx)
	}
	go TKLog(&Log{
		Uid:       uid,
		Tag:       logger.tag,
		LogModule: GameLogModule,
		TracedId:  traceId,
		LogLevel:  Info,
		Service:   logger.Service,
		Content:   content,
		ExtField:  ex,
	})
}

// TKLogError 打印Error日志 LogModule为game
func TKLogError(ctx context.Context, uid int64, content string, extra ...map[string]interface{}) {
	var l = Log{
		Uid:       uid,
		LogModule: GameLogModule,
		TracedId:  nil,
		LogLevel:  Error,
		Service:   logger.Service,
		Content:   content,
		Code:      0,
		Msg:       "",
		ExtField:  make(map[string]interface{}),
		Tag:       logger.tag,
		Cost:      0,
	}

	if l.Uid == 0 {
		l.Uid = GetUidFromCtx(ctx)
	}
	if traceId := GetTraceId(ctx); traceId != "" {
		l.TracedId = traceId
	}
	if l.TracedId == nil && l.Uid == 0 {
		l.ExtField["stack"] = string(debug.Stack())
	}

	if len(extra) != 0 {
		l.ExtField = extra[0]
	}

	go TKLog(&l)
}

// TkLogSystemError 打印系统级别日志 LogModule为system
func TkLogSystemError(ctx context.Context, uid int64, content string, extra ...map[string]interface{}) {
	var ex map[string]interface{}
	var traceId string
	traceId = GetTraceId(ctx)
	if uid == 0 {
		uid = GetUidFromCtx(ctx)
	}
	if len(extra) != 0 {
		ex = extra[0]
	}

	go TKLog(&Log{
		Uid:       uid,
		TracedId:  traceId,
		LogModule: SystemLogModule,
		Tag:       logger.tag,
		LogLevel:  Error,
		Service:   logger.Service,
		Content:   content,
		ExtField:  ex,
	})
}

func TkLogSystemInfo(ctx context.Context, uid int64, content string, extra ...map[string]interface{}) {
	var ex map[string]interface{}
	if len(extra) != 0 {
		ex = extra[0]
	}
	var traceId string
	if uid == 0 {
		uid = GetUidFromCtx(ctx)
	}
	traceId = GetTraceId(ctx)

	go TKLog(&Log{
		Uid:       uid,
		TracedId:  traceId,
		LogModule: SystemLogModule,
		Tag:       logger.tag,
		LogLevel:  Info,
		Service:   logger.Service,
		Content:   content,
		ExtField:  ex,
	})
}

func Close() {
	if logger != nil && logger.httpClient != nil {
		logger.httpClient.CloseIdleConnections()
	}
}

func GetTraceId(ctx context.Context) string {
	if ctx == nil {
		return ""
	}
	meta, ok := metadata.FromContext(ctx)
	if ok {
		tmps := strings.Split(meta["Uber-Trace-Id"], ":")
		if len(tmps) != 0 {
			return tmps[0]
		}
	}
	return ""
}
